#include "dirselector.h"

DirSelector* DirSelector::s_instance;

DirSelector::DirSelector(QWidget * parent, const char * name)
  :QListView(parent, name)
{
  s_instance = this;
  
  addColumn(QString::null);
  setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
  setResizeMode(QListView::LastColumn);
  reinterpret_cast<QWidget*>(header())->hide();
}

DirSelector::~DirSelector()
{

}

void DirSelector::setDirectories(const QStringList & l)
{
  fDirectories = l;
}

void DirSelector::setTopDir(const KURL & url)
{
  if (fTopDir!=url) {
    fDirectories.clear();
  }
  fTopDir = url;
}

void DirSelector::refresh()
{
  clear();
  new DirSelectorItem(this, fTopDir);
}

DirSelectorItem::DirSelectorItem(QListView * parent, const KURL & url)
  :QCheckListItem(parent, "/", QCheckListItem::CheckBox)
  ,fDirLister(true)
  ,fURL(url)
  ,fListed(false)
  ,fDisabled(true)
  ,fRDisabled(false)
{
//   fDirLister.setDirOnlyMode(true);

  connect(&fDirLister, SIGNAL(newItems(const KFileItemList &)), this, SLOT(newItems(const KFileItemList &)));
  setOpen(true);
  setVisible(true);
}

DirSelectorItem::DirSelectorItem(QListViewItem * parent, const KURL & url, bool disable = false, bool rdisable = false)
  :QCheckListItem(parent, url.fileName(), QCheckListItem::CheckBox)
  ,fDirLister(true)
  ,fURL(url)
  ,fListed(false)
  ,fDisabled(disable)
  ,fRDisabled(rdisable)
{
//   fDirLister.setDirOnlyMode(true);

//   setExpandable(true);
  connect(&fDirLister, SIGNAL(newItems(const KFileItemList &)), this, SLOT(newItems(const KFileItemList &)));
  connect(&fDirLister, SIGNAL(completed()), this, SLOT(completed()));
  connect(&fDirLister, SIGNAL(canceled()), this, SLOT(completed()));
}

QCheckListItem * DirSelectorItem::parent() const
{
  return (QCheckListItem*) QListViewItem::parent();
}

void DirSelectorItem::setOpen(bool b)
{
  if (!fListed) {
    fDirLister.openURL(fURL, true);
    fListed = true;
  }

  QListViewItem::setOpen(b);
}

void DirSelectorItem::stateChange(bool b)
{
  if(isRDisabled()) return;
  
  for( QListViewItem *item = firstChild(); item; item = item->nextSibling() )
    if ( dynamic_cast<DirSelectorItem*>( item ) && !dynamic_cast<DirSelectorItem*>( item )->isRDisabled() )
          static_cast<QCheckListItem*>(item)->QCheckListItem::setOn( b );

  if(isDisabled()) return;
  
  DirSelector * dirs = (DirSelector*) listView();
  QStringList::Iterator it = dirs->directories().find(fURL.path());
  if (isOn()) {
    if (it==dirs->directories().end()) {
      dirs->directories() << fURL.path();
    }
    
    // Deselect subdirectories
    QStringList::Iterator diriter = dirs->directories().begin();
    while ( diriter != dirs->directories().end() )
    {
        // Since the dir "/" starts with '/', we need a hack to stop it removing
        // itself (it being the only path with a trailing '/')
        if ( (*diriter).startsWith( fURL.path(1) ) && *diriter != "/" )
            diriter = dirs->directories().erase(diriter);
        else
            ++diriter;
    }
  } else {
    if ( it != dirs->directories().end() )
      dirs->directories().erase(it);
    
    QStringList::Iterator diriter = dirs->directories().begin();
    while ( diriter != dirs->directories().end() )
    {
        if ( (*diriter).startsWith( fURL.path(1) ) )   //path(1) adds a trailing '/'
        {
            diriter = dirs->directories().erase(diriter);
        }
        else
            ++diriter;
    }
  }

  listView()->triggerUpdate();
}

void DirSelectorItem::activate()
{
  if(!isDisabled())
    QCheckListItem::activate();
}

void DirSelectorItem::newItems(const KFileItemList & list)
{
  DirSelector * dirs = (DirSelector*) listView();

  for (KFileItemListIterator it(list);*it;++it) {
    KURL url = (*it)->url();
    bool disable = false;
    bool rdisable = this->isRDisabled();
    if( DirSelector::instance()->pathsDisabled.contains( ( *it )->url().path() ) ) disable = true;
    if( DirSelector::instance()->pathsRDisabled.contains( ( *it )->url().path() ) ) rdisable = true;
    DirSelectorItem *item = new DirSelectorItem(this, url, disable, rdisable);
    
    if(! item->isRDisabled()) item->setOn(isOn() || dirs->directories().contains(item->url().path()));
    item->setPixmap( 0, (*it)->pixmap( KIcon::SizeSmall ) );
    item->setExpandable((*it)->isDir());
  }
}

void DirSelectorItem::paintCell(QPainter * p, const QColorGroup & cg, int column, int width, int align)
{
  DirSelector * dirs = (DirSelector*) listView();

  bool dirty = false;

  for (uint i=0;i<dirs->directories().count();i++) {
    if (dirs->directories()[i].startsWith(fURL.path())) dirty = true;
  }

  QColorGroup _cg = cg;
  if (dirty) _cg.setColor(QColorGroup::Text, listView()->colorGroup().link());

  QCheckListItem::paintCell(p, isDisabled() ? listView()->palette().disabled() : _cg, column, width, align);
}

void DirSelectorItem::completed()
{
  if (childCount()==0) {
    setExpandable(false);
    repaint();
  }
}
